from channels.routing import route, route_class
from .consumers import ws_message, ws_connect, ws_disconnect
from .registry import DEMULTIPLEXERS
from .registry import EXTRA_ROUTES

channel_routing = [
    route("websocket.connect", ws_connect, path=r"^/tv/$"),
    route("websocket.receive", ws_message, path=r"^/tv/$"),
    route("websocket.disconnect", ws_disconnect, path=r"^/tv/$"),
]

for Demultiplexer in DEMULTIPLEXERS:
    channel_routing.append(
        route_class(Demultiplexer)
    )

for route in EXTRA_ROUTES:
    channel_routing.append(route)



from functools import wraps
from pydoc import locate

from django.conf import settings
from django.dispatch import receiver
from django.test.signals import setting_changed
from django.db import connection

from channels.binding.base import BindingMetaclass
from channels.binding.websockets import WebsocketBindingWithMembers
from channels.generic.websockets import WebsocketDemultiplexer

from television.registry import LISTENERS, DEMULTIPLEXERS
from television.exceptions import ListenerNotFound


def require_auth(fun):
    @wraps(fun)
    def _inner(message, *args, **kwargs):
        if not message.user.is_authenticated():
            raise Exception("NOAUTH")
        return fun(message, *args, **kwargs)
    return _inner


def require_staff(fun):
    @wraps(fun)
    def _inner(message, *args, **kwargs):
        if not message.user.is_staff:
            raise Exception("NOSTAFF")
        return fun(message, *args, **kwargs)
    return _inner


def require_superuser(fun):
    @wraps(fun)
    def _inner(message, *args, **kwargs):
        if not message.user.is_superuser:
            raise Exception("NOSUDO")
        return fun(message, *args, **kwargs)
    return _inner


APPS_FINISHED_LOADING = False
def add_listener(channel):
    def decorator(fun):
        @wraps(fun)
        def _inner(*args, **kwargs):
            return fun(*args, **kwargs)
        if channel in LISTENERS:
            if not APPS_FINISHED_LOADING: # prevent test runner from blowing up with exception 'Listener already registered...'
                raise Exception(f"Listener already registered for channel: {channel}")
        module_value = "%s.%s" % (fun.__module__, fun.__name__)
        if module_value.split('.')[0] == 'television':
            # is OK to map anonymous function created in this library
            LISTENERS[channel] = fun
        elif module_value in LISTENERS.values():
            # throw helpful error explaining function name is defined twice
            raise Exception(f"Listener callback function defined twice: {module_value}")
        else:
            LISTENERS[channel] = module_value
        return _inner
    return decorator


def call_listener(channel, message, *args, **kwargs):
    try:
        listener = LISTENERS[channel]
    except KeyError:
        raise ListenerNotFound(f"Listener not defined for channel '{channel}'")
    if not callable(listener):
        listener = locate(listener)  # may be of form: 'module.path.function'
    return listener(message, *args, **kwargs)


def add_data_binding_staff(fields='__all__', send_members=[], exclude=[]):
    def decorate(model_class):
        nonlocal fields

        send_members.append('pk')  # always send pk

        if fields == '__all__':
            fields = [x.name for x in model_class._meta.fields]
        elif fields == None:
            fields = []

        class StaffAuthModelBinding(WebsocketBindingWithMembers):
            model = model_class
            stream = 'television-updates'

            @classmethod
            def group_names(cls, instance):
                return ['staff']

            def has_permission(self, user, action, pk):
                return user.is_staff

        setattr(StaffAuthModelBinding, 'fields', fields)
        setattr(StaffAuthModelBinding, 'send_members', send_members)
        setattr(StaffAuthModelBinding, 'exclude', exclude)

        class Demultiplexer(WebsocketDemultiplexer):
            consumers = {
                'television-updates': StaffAuthModelBinding.consumer,
            }

            def connection_groups(self):
                return ["staff"]

        DEMULTIPLEXERS.append(Demultiplexer)

        # add 'app.model.list' endpoint for serialized objects
        serialize_data = StaffAuthModelBinding().serialize_data
        listener_name = f'{model_class._meta.label_lower}.list'
        @require_staff
        @add_listener(listener_name)
        def listener(message):
            queryset = model_class.objects.all().order_by('-id')
            result = []
            for instance in queryset:
                result.append(serialize_data(instance))
            return result

        return model_class
    return decorate


def add_data_binding_superuser(fields='__all__', send_members=[], exclude=[]):
    def decorate(model_class):
        nonlocal fields

        send_members.append('pk')  # always send pk

        if fields == '__all__':
            fields = [x.name for x in model_class._meta.fields]
        elif fields == None:
            fields = []

        class SuperuserAuthModelBinding(WebsocketBindingWithMembers):
            model = model_class
            stream = 'television-updates'

            @classmethod
            def group_names(cls, instance):
                return ['superusers']

            def has_permission(self, user, action, pk):
                return user.is_superuser

        setattr(SuperuserAuthModelBinding, 'fields', fields)
        setattr(SuperuserAuthModelBinding, 'send_members', send_members)
        setattr(SuperuserAuthModelBinding, 'exclude', exclude)

        class Demultiplexer(WebsocketDemultiplexer):
            consumers = {
                'television-updates': SuperuserAuthModelBinding.consumer,
            }

            def connection_groups(self):
                return ["superusers"]

        DEMULTIPLEXERS.append(Demultiplexer)

        # add 'app.model.list' endpoint for serialized objects
        serialize_data = SuperuserAuthModelBinding().serialize_data
        listener_name = f'{model_class._meta.label_lower}.list'
        @require_superuser
        @add_listener(listener_name)
        def listener(message):
            queryset = model_class.objects.all().order_by('-id')
            result = []
            for instance in queryset:
                result.append(serialize_data(instance))
            return result

        return model_class
    return decorate


def add_data_binding_owner(fields='__all__', send_members=[], exclude=[], owner_field='user'):
    raise NotImplementedError('To do.')

    def decorate(model_class):
        nonlocal fields

        send_members.append('pk')  # always send pk

        if fields == '__all__':
            fields = [x.name for x in model_class._meta.fields]
        elif fields == None:
            fields = []

        class OwnerAuthModelBinding(WebsocketBindingWithMembers):
            model = model_class
            stream = 'television-updates'

            @classmethod
            def group_names(cls, instance):
                owner = instance
                for attr in cls.owner_field.split('.'):
                    owner = getattr(owner, attr)
                return ["users", f'users.{owner.id}']

            def has_permission(self, user, action, pk):
                raise NotImplementedError('To do.')

        setattr(OwnerAuthModelBinding, 'fields', fields)
        setattr(OwnerAuthModelBinding, 'send_members', send_members)
        setattr(OwnerAuthModelBinding, 'exclude', exclude)
        setattr(OwnerAuthModelBinding, 'owner_field', owner_field)

        class Demultiplexer(WebsocketDemultiplexer):
            consumers = {
                'television-updates': OwnerAuthModelBinding.consumer,
            }

            def connection_groups(self):
                return ["users"]

        DEMULTIPLEXERS.append(Demultiplexer)

        # add 'app.model.list' endpoint for serialized objects
        serialize_data = SuperuserAuthModelBinding().serialize_data
        listener_name = f'{model_class._meta.label_lower}.list'
        @require_auth
        @add_listener(listener_name)
        def listener(message):
            raise NotImplementedError('To do.')

        return model_class
    return decorate



from channels import Group
import json
import hashlib
import datetime


def timestamp():
    return datetime.datetime.now().strftime('%Y-%b-%d %H:%M:%S.%f')[:-3]


def send_to_group(group, channel, payload):
    return Group(group).send({'text': json.dumps({'stream': channel, 'payload': payload})})


def staff_log(message):
    return send_to_group('staff', 'staff.log', {'message': message})



from django.apps import AppConfig, apps
import importlib
import traceback

from .registry import LISTENERS

class AppConfig(AppConfig): # Our app config class
    name = 'television'
    verbose_name = 'Television'

    def ready(self):
        # 1) Register all listeners
        for app in apps.app_configs.keys():
            app_config = apps.app_configs[app]
            module_name = f'{app_config.module.__name__}.listeners'
            try:
                importlib.import_module(module_name)
            except ModuleNotFoundError as ex:
                if 'listeners' not in ex.name:
                    raise # import failed within module.listeners source
                pass  # This means module.listeners does not exist.
            except ImportError:
                raise  # Syntax/other error. Raise.
        if LISTENERS:
            print('\n[django-television]')
            for listener, fn in LISTENERS.items():
                print(f'\t* {listener}: {fn}')
            print('\n')
        else:
            print('\n[django-television] No listeners.py found for INSTALLED_APPS modules.')

        import television.decorators
        television.decorators.APPS_FINISHED_LOADING = True


from setuptools import setup
from television import __version__

from setuptools.command.develop import develop
from setuptools.command.install import install


class DevelopCommand(develop):
    def run(self):
        assert sys.version_info >= (3, 6)  # python3.6+
        develop.run(self)


class InstallCommand(install):
    def run(self):
        assert sys.version_info >= (3, 6)  # python3.6+
        install.run(self)


setup(
    name='django-television',
    version=__version__,
    packages=['television'],
    scripts=[],
    install_requires=[
        'Django>=1.11',
        'channels==1.1.8'
    ],
    cmdclass={
        'develop': DevelopCommand,
        'install': InstallCommand,
    },
    author='Patrick Paul',
    author_email='patrick@astrohaus.com',
    description='Back-end Python utilities for use with Django Channels',
    license='MIT',
    # keywords='',
    url='https://github.com/pztrick/django-television/',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
    ],
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

__version__ = "0.1.4"

default_app_config = 'television.app.AppConfig'

try:
    from .decorators import add_listener, call_listener, require_auth, require_staff, require_superuser, add_data_binding_staff, add_data_binding_superuser, add_data_binding_owner
except ModuleNotFoundError:
    pass  # prevent setuptools from exploding during packaging



import json
import traceback
from django.http import HttpResponse
from channels.handler import AsgiHandler
from channels import Group
from .decorators import add_listener, call_listener
from django.conf import settings
from django.db import connection
from channels.auth import http_session_user, channel_session_user, channel_session_user_from_http
from .utils import timestamp

@channel_session_user_from_http
def ws_connect(message):
    if message.user.is_authenticated:
        Group(f"users").add(message.reply_channel)
        Group(f"users.{message.user.id}").add(message.reply_channel)
        print(f"ws_connect: added to group 'users'")
    if message.user.is_superuser:
        Group("superusers").add(message.reply_channel)
        print("ws_connect: added to group 'superusers'")
    if message.user.is_staff:
        Group("staff").add(message.reply_channel)
        print("ws_connect: added to group 'staff'")
    Group("chat").add(message.reply_channel)
    message.reply_channel.send({"accept": True})

@channel_session_user
def ws_message(message):
    try:
        request = json.loads(message['text'])
        replyTo = request.get('replyTo', None)
        channel = request['channel']
        payload = request.get('payload', [])
        if settings.DEBUG:
            print(f"[{timestamp()}] ws_message received on channel '{channel}'")
        n_queries = len(connection.queries)
        result = call_listener(channel, message, *payload)
        n_queries = len(connection.queries) - n_queries
        response = {
            'replyTo': replyTo,
            'payload': result
        }
        message.reply_channel.send({
            'text': json.dumps(response, default=str)
        })
        if settings.DEBUG:
            print(f"[{timestamp()}] ws_message replied to on channel '{channel}' ({n_queries} SQL queries)")

    except Exception as ex:
        print(traceback.format_exc())
        try:
            errorTo = request.get('errorTo', None)
        except:
            errorTo = None
        formatted_lines = traceback.format_exc().splitlines()
        if settings.DEBUG:
            result = "Backend Error\n%s\n%s" % (formatted_lines[1], formatted_lines[-1])
        else:
            result = "Backend Error\n%s" % (formatted_lines[-1], )
        response = {
            'replyTo': errorTo,
            'payload': result
        }
        message.reply_channel.send({
            'text': json.dumps(response, default=str)
        })
        raise ex

@channel_session_user
def ws_disconnect(message):
    if message.user.is_authenticated:
        print("ws_disconnect: %s" % message.user.email)
        Group(f"users").discard(message.reply_channel)
        Group(f"users.{message.user.id}").discard(message.reply_channel)
    if message.user.is_superuser:
        Group("superusers").discard(message.reply_channel)
    if message.user.is_staff:
        Group("staff").discard(message.reply_channel)
    Group("chat").discard(message.reply_channel)
